from typing import Optional

import rhinoscriptsyntax as rs  # type: ignore
from compas_skeleton.datastructures import Skeleton
from compas_triangle.delaunay import conforming_delaunay_triangulation
from compas_triangle.rhino import discretise_boundary
from compas_triangle.rhino import discretise_constraints

import compas_rhino
import compas_rhino.conversions
import compas_rhino.objects
from compas.geometry import NurbsCurve
from compas.geometry import Point
from compas_rv.datastructures import Pattern


def make_pattern_from_rhinolines() -> Optional[Pattern]:
    guids = compas_rhino.objects.select_lines("Select lines")
    if not guids:
        return

    lines = compas_rhino.objects.get_line_coordinates(guids)
    if not lines:
        return

    pattern: Pattern = Pattern.from_lines(lines, delete_boundary_face=True)  # type: ignore

    rs.HideObjects(guids)

    return pattern


def make_pattern_from_rhinomesh() -> Optional[Pattern]:
    guid = compas_rhino.objects.select_mesh("Select a mesh")
    if not guid:
        return

    obj = compas_rhino.objects.find_object(guid)
    pattern: Pattern = compas_rhino.conversions.mesh_to_compas(obj.Geometry, cls=Pattern)  # type: ignore

    rs.HideObject(guid)

    return pattern


def make_pattern_from_rhinosurface() -> Optional[Pattern]:
    guid = compas_rhino.objects.select_surface("Select a surface")
    if not guid:
        return

    U = rs.GetInteger(message="U faces", number=16, minimum=2, maximum=64)
    if not U:
        return

    V = rs.GetInteger(message="V faces", number=4, minimum=2, maximum=64)
    if not V:
        return

    obj = compas_rhino.objects.find_object(guid)
    brep = obj.Geometry
    surface = brep.Surfaces[0]
    pattern: Pattern = compas_rhino.conversions.surface_to_compas_mesh(surface, nu=U, nv=V, weld=True, cls=Pattern)  # type: ignore

    rs.HideObject(guid)

    return pattern


def make_pattern_from_meshgrid() -> Optional[Pattern]:
    DX = rs.GetInteger(message="X Size", number=10)
    if not DX:
        return

    DY = rs.GetInteger(message="Y Size", number=DX)
    if not DY:
        return

    NX = rs.GetInteger(message="Number of faces in X", number=10)
    if not NX:
        return

    NY = rs.GetInteger(message="Number of faces in Y", number=NX)
    if not NY:
        return

    pattern: Pattern = Pattern.from_meshgrid(dx=DX, nx=NX, dy=DY, ny=NY)  # type: ignore
    return pattern


def make_pattern_from_triangulation() -> Optional[Pattern]:
    boundary_guids = compas_rhino.objects.select_curves("Select outer boundary.")
    if not boundary_guids:
        return

    rs.UnselectAllObjects()
    hole_guids = compas_rhino.objects.select_curves("Select inner boundaries.")

    rs.UnselectAllObjects()
    segments_guids = compas_rhino.objects.select_curves("Select constraint curves.")

    rs.UnselectAllObjects()

    target_length = rs.GetReal("Specifiy target edge length.", 1.0)
    if not target_length:
        return

    boundary = discretise_boundary(boundary_guids, target_length)
    holes = None
    segments = None
    curves = None  # type: ignore

    if hole_guids:
        holes = discretise_constraints(hole_guids, target_length)

    if segments_guids:
        segments = discretise_constraints(segments_guids, target_length)
        curves: list[NurbsCurve] = [NurbsCurve.from_interpolation(segment) for segment in segments]

    points, triangles = conforming_delaunay_triangulation(
        boundary,
        polylines=segments,
        polygons=holes,
        area=target_length**2 / 2,
    )
    pattern = Pattern.from_vertices_and_faces(points, triangles)

    fixed = [vertex for boundary in pattern.vertices_on_boundaries() for vertex in boundary]
    if curves:
        for index, point in enumerate(points):
            for curve in curves:
                closest: Point = curve.closest_point(point)
                if closest.distance_to_point(point) < 0.1 * target_length:
                    fixed.append(index)

    pattern.smooth_area(fixed=fixed)


def make_pattern_from_skeleton() -> Optional[Pattern]:
    guids = compas_rhino.objects.select_lines("Select skeleton lines.")
    if not guids:
        return

    rs.UnselectAllObjects()

    width = rs.GetReal("Specifiy skeleton width.", 1.0)
    if not width:
        return

    angle = rs.GetReal("Specifiy skeleton leaf angle (degrees).", 30)
    if not angle:
        return

    density = rs.GetInteger("Specifiy skeleton density.", 2)
    if not density:
        return

    objects = [compas_rhino.objects.find_object(guid) for guid in guids]
    curves = [obj.Geometry for obj in objects]
    lines = [compas_rhino.conversions.curve_to_compas_line(curve) for curve in curves]

    skeleton = Skeleton(lines)
    skeleton.node_width = width
    skeleton.leaf_angle = angle
    skeleton.density = density
    pattern = skeleton.pattern.copy(cls=Pattern)

    return pattern



import math

from compas.geometry import distance_point_point_xy
from compas.geometry import mirror_points_line
from compas.geometry import rotate_points_xy
from compas_rv.datastructures import Pattern


def mirror_4x(line, line_hor, line_ver, lines):
    """Helper to mirror an object 4 times."""
    lines.append(line)
    a_mirror, b_mirror = mirror_points_line(line, line_hor)
    lines.append([a_mirror, b_mirror])
    a_mirror, b_mirror = mirror_points_line([a_mirror, b_mirror], line_ver)
    lines.append([a_mirror, b_mirror])
    a_mirror, b_mirror = mirror_points_line(line, line_ver)
    lines.append([a_mirror, b_mirror])
    return lines


def mirror_8x(line, origin, line_hor, line_ver, lines):
    """Helper to mirror an object 8 times."""
    lines = mirror_4x(line, line_hor, line_ver, lines)
    rot = rotate_points_xy(line, math.pi / 2, origin=origin)
    lines = mirror_4x(rot, line_hor, line_ver, lines)
    return lines


def append_mirrored_lines(line, list_, line_hor, line_ver):
    """Helper to mirror an object 8 times and add to the list"""
    mirror_a = mirror_points_line(line, line_hor)
    mirror_b = mirror_points_line(mirror_a, line_ver)
    mirror_c = mirror_points_line(line, line_ver)
    list_.append(mirror_a)
    list_.append(mirror_b)
    list_.append(mirror_c)


def create_cross_pattern(
    xy_span=[[0.0, 10.0], [0.0, 10.0]],
    discretisation=10,
) -> Pattern:
    """Construct a FormDiagram based on cross discretiastion with orthogonal arrangement and diagonal.

    Parameters
    ----------
    xy_span : list, optional
        List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
    discretisation : int, optional
        Set the density of the grid in x and y directions, by default 10
    fix : str, optional
        Option to select the constrained nodes: 'corners', 'all' are accepted., by default 'corners'

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.

    Notes
    ----------------------
    Position of the quadrants is as in the schema below:

        Q3
    Q2      Q1
        Q4
    """
    if isinstance(discretisation, list):
        discretisation = discretisation[0]

    y1 = float(xy_span[1][1])
    y0 = float(xy_span[1][0])
    x1 = float(xy_span[0][1])
    x0 = float(xy_span[0][0])
    x_span = x1 - x0
    y_span = y1 - y0
    dx = x_span / discretisation
    dy = y_span / discretisation

    lines = []

    for i in range(discretisation + 1):
        for j in range(discretisation + 1):
            if i < discretisation and j < discretisation:
                # Vertical Members:
                xa = x0 + dx * i
                ya = y0 + dy * j
                xb = x0 + dx * (i + 1)
                yb = y0 + dy * j
                # Horizontal Members:
                xc = x0 + dx * i
                yc = y0 + dy * j
                xd = x0 + dx * i
                yd = y0 + dy * (j + 1)
                lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])
                lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
                if i == j:
                    # Diagonal Members in + Direction:
                    xc = x0 + dx * i
                    yc = y0 + dy * j
                    xd = x0 + dx * (i + 1)
                    yd = y0 + dy * (j + 1)
                    lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
                if i + j == discretisation:
                    # Diagonal Members in - Direction:
                    xc = x0 + dx * i
                    yc = y0 + dy * j
                    xd = x0 + dx * (i - 1)
                    yd = y0 + dy * (j + 1)
                    lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
                    if i == (discretisation - 1):
                        xc = x0 + dx * i
                        yc = y0 + dy * j
                        xd = x0 + dx * (i + 1)
                        yd = y0 + dy * (j - 1)
                        lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
            else:
                if i == discretisation and j < discretisation:
                    # Vertical Members on last column:
                    xa = x0 + dx * j
                    ya = y0 + dy * i
                    xb = x0 + dx * (j + 1)
                    yb = y0 + dy * i
                    # Horizontal Members:
                    xc = x0 + dx * i
                    yc = y0 + dy * j
                    xd = x0 + dx * i
                    yd = y0 + dy * (j + 1)
                    lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])
                    lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])

    pattern: Pattern = Pattern.from_lines(lines, delete_boundary_face=True)  # type: ignore
    return pattern


def create_cross_diagonal_pattern(
    xy_span=[[0.0, 10.0], [0.0, 10.0]],
    partial_bracing_modules=None,
    discretisation=10,
) -> Pattern:
    """Construct a FormDiagram based on a mixture of cross and fan discretiastion

    Parameters
    ----------
    xy_span : list, optional
        List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
    discretisation : int, optional
        Set the density of the grid in x and y directions, by default 10
    partial_bracing_modules : str, optional
        If partial bracing modules are included, by default None
    fix : str, optional
        Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.

    """
    if isinstance(discretisation, list):
        discretisation = discretisation[0]

    y1 = float(xy_span[1][1])
    y0 = float(xy_span[1][0])
    x1 = float(xy_span[0][1])
    x0 = float(xy_span[0][0])
    x_span = x1 - x0
    y_span = y1 - y0
    dx = x_span / discretisation
    dy = y_span / discretisation

    xc0 = x0 + x_span / 2
    yc0 = y0 + y_span / 2

    nx = ny = int(discretisation / 2)
    if partial_bracing_modules is None:
        nstop = 0
    else:
        nstop = nx - partial_bracing_modules  # Test to stop

    line_hor = [[x0, yc0, 0.0], [xc0, yc0, 0.0]]
    line_ver = [[xc0, y0, 0.0], [xc0, yc0, 0.0]]
    origin = [xc0, yc0, 0.0]

    lines = []

    for i in range(nx):
        for j in range(ny + 1):
            if j <= i:
                if i >= nstop and j >= nstop:
                    # Diagonal Members:
                    xa = x0 + dx * i
                    ya = y0 + dy * 1 * (i - j) / (nx - j) + dy * j
                    xb = x0 + dx * (i + 1)
                    yb = y0 + dy * 1 * (i - j + 1) / (nx - j) + dy * j
                    lin = [[xa, ya, 0.0], [xb, yb, 0.0]]
                    lines = mirror_8x(lin, origin, line_hor, line_ver, lines)

                if i == j and i < nx - 1:
                    # Main diagonal:
                    xa = x0 + dx * i
                    ya = y0 + dy * j
                    xb = x0 + dx * (i + 1)
                    yb = y0 + dy * (j + 1)
                    lin = [[xa, ya, 0.0], [xb, yb, 0.0]]
                    lines = mirror_4x(lin, line_hor, line_ver, lines)

                # Horizontal Members:
                xa = x0 + dx * i
                ya = y0 + dy * j
                xb = x0 + dx * (i + 1)
                yb = y0 + dy * j
                lin = [[xa, ya, 0.0], [xb, yb, 0.0]]
                lines = mirror_8x(lin, origin, line_hor, line_ver, lines)

                i += 1
                # Vertical Members:
                xa = x0 + dx * i
                ya = y0 + dy * j
                xb = x0 + dx * i
                yb = y0 + dy * (j + 1)

                if i >= nstop and j >= nstop:
                    x_ = xa
                    y_ = y0 + dy * 1 * (i - j) / (nx - j) + dy * j
                    lin = [[xa, ya, 0.0], [x_, y_, 0.0]]
                    lines = mirror_8x(lin, origin, line_hor, line_ver, lines)
                    lin = [[x_, y_, 0.0], [xb, yb, 0.0]]
                    lines = mirror_8x(lin, origin, line_hor, line_ver, lines)
                else:
                    lin = [[xa, ya, 0.0], [xb, yb, 0.0]]
                    lines = mirror_8x(lin, origin, line_hor, line_ver, lines)

                i -= 1

    pattern: Pattern = Pattern.from_lines(lines, delete_boundary_face=True)  # type: ignore
    return pattern


def create_cross_with_diagonal_pattern(
    xy_span=[[0.0, 10.0], [0.0, 10.0]],
    discretisation=10,
) -> Pattern:
    """Construct a FormDiagram based on cross discretiastion with diagonals.

    Parameters
    ----------
    xy_span : list, optional
        List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
    discretisation : int, optional
        Set the density of the grid in x and y directions, by default 10
    fix : str, optional
        Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.
    """

    if isinstance(discretisation, list):
        discretisation = discretisation[0]

    y1 = float(xy_span[1][1])
    y0 = float(xy_span[1][0])
    x1 = float(xy_span[0][1])
    x0 = float(xy_span[0][0])
    x_span = x1 - x0
    y_span = y1 - y0
    dx = x_span / discretisation
    dy = y_span / discretisation

    lines = []

    for i in range(discretisation + 1):
        for j in range(discretisation + 1):
            if i < discretisation and j < discretisation:
                # Hor Members:
                xa = x0 + dx * i
                ya = y0 + dy * j
                xb = x0 + dx * (i + 1)
                yb = y0 + dy * j
                # Ver Members:
                xc = x0 + dx * i
                yc = y0 + dy * j
                xd = x0 + dx * i
                yd = y0 + dy * (j + 1)
                lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])
                lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
                # lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
                if (i < discretisation / 2 and j < discretisation / 2) or (i >= discretisation / 2 and j >= discretisation / 2):
                    # Diagonal Members in + Direction:
                    xc = x0 + dx * i
                    yc = y0 + dy * j
                    xd = x0 + dx * (i + 1)
                    yd = y0 + dy * (j + 1)
                    lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
                else:
                    # Diagonal Members in - Direction:
                    xc = x0 + dx * i
                    yc = y0 + dy * (j + 1)
                    xd = x0 + dx * (i + 1)
                    yd = y0 + dy * j
                    lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
                    # if i == (discretisation - 1):
                    #     xc = x0 + dx*i
                    #     yc = y0 + dy*j
                    #     xd = x0 + dx*(i + 1)
                    #     yd = y0 + dy*(j - 1)
                    #     lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
            else:
                if i == discretisation and j < discretisation:
                    # Vertical Members on last column:
                    xa = x0 + dx * j
                    ya = y0 + dy * i
                    xb = x0 + dx * (j + 1)
                    yb = y0 + dy * i
                    # Horizontal Members:
                    xc = x0 + dx * i
                    yc = y0 + dy * j
                    xd = x0 + dx * i
                    yd = y0 + dy * (j + 1)
                    lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])
                    lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])

    pattern: Pattern = Pattern.from_lines(lines, delete_boundary_face=True)  # type: ignore
    return pattern


def create_fan_pattern(
    xy_span=[[0.0, 10.0], [0.0, 10.0]],
    discretisation=[10, 10],
) -> Pattern:
    """Helper to construct a FormDiagram based on fan discretiastion with straight lines to the corners.

    Parameters
    ----------
    xy_span : list, optional
        List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
    discretisation : int, optional
        Set the density of the grid in x and y directions, by default 10
    fix : str, optional
        Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.
    """

    if isinstance(discretisation, int):
        discretisation = [discretisation, discretisation]
    if discretisation[0] % 2 != 0 or discretisation[1] % 2 != 0:
        msg = "Warning!: discretisation of this form diagram has to be even."
        raise ValueError(msg)

    y1 = float(xy_span[1][1])
    y0 = float(xy_span[1][0])
    x1 = float(xy_span[0][1])
    x0 = float(xy_span[0][0])

    x_span = x1 - x0
    y_span = y1 - y0
    xc0 = x0 + x_span / 2
    yc0 = y0 + y_span / 2
    division_x = discretisation[0]
    division_y = discretisation[1]
    dx = float(x_span / division_x)
    dy = float(y_span / division_y)
    nx = int(division_x / 2)
    ny = int(division_y / 2)
    line_hor = [[x0, yc0, 0.0], [xc0, yc0, 0.0]]
    line_ver = [[xc0, y0, 0.0], [xc0, yc0, 0.0]]

    lines = []

    for i in range(nx):
        for j in range(ny + 1):
            # Diagonal Members:
            xa = x0 + dx * i
            ya = y0 + dy * j * i / nx
            xb = x0 + dx * (i + 1)
            yb = y0 + dy * j * (i + 1) / nx
            lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

            a_mirror, b_mirror = mirror_points_line([[xa, ya, 0.0], [xb, yb, 0.0]], line_hor)
            lines.append([a_mirror, b_mirror])
            a_mirror, b_mirror = mirror_points_line([a_mirror, b_mirror], line_ver)
            lines.append([a_mirror, b_mirror])
            a_mirror, b_mirror = mirror_points_line([[xa, ya, 0.0], [xb, yb, 0.0]], line_ver)
            lines.append([a_mirror, b_mirror])

            xa_ = x0 + dx * j * i / nx
            ya_ = y0 + dy * i
            xb_ = x0 + dx * j * (i + 1) / nx
            yb_ = y0 + dy * (i + 1)
            lines.append([[xa_, ya_, 0.0], [xb_, yb_, 0.0]])

            a_mirror, b_mirror = mirror_points_line([[xa_, ya_, 0.0], [xb_, yb_, 0.0]], line_hor)
            lines.append([a_mirror, b_mirror])
            a_mirror, b_mirror = mirror_points_line([a_mirror, b_mirror], line_ver)
            lines.append([a_mirror, b_mirror])
            a_mirror, b_mirror = mirror_points_line([[xa_, ya_, 0.0], [xb_, yb_, 0.0]], line_ver)
            lines.append([a_mirror, b_mirror])

            if j < ny:
                # Vertical or Horizontal Members:
                xc = x0 + dx * (i + 1)
                yc = y0 + dy * j * (i + 1) / nx
                xd = x0 + dx * (i + 1)
                yd = y0 + dy * (j + 1) * (i + 1) / nx
                lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])

                c_mirror, d_mirror = mirror_points_line([[xc, yc, 0.0], [xd, yd, 0.0]], line_hor)
                lines.append([c_mirror, d_mirror])
                c_mirror, d_mirror = mirror_points_line([c_mirror, d_mirror], line_ver)
                lines.append([c_mirror, d_mirror])
                c_mirror, d_mirror = mirror_points_line([[xc, yc, 0.0], [xd, yd, 0.0]], line_ver)
                lines.append([c_mirror, d_mirror])

                xc_ = x0 + dx * j * (i + 1) / nx
                yc_ = y0 + dy * (i + 1)
                xd_ = x0 + dx * (j + 1) * (i + 1) / nx
                yd_ = y0 + dy * (i + 1)
                lines.append([[xc_, yc_, 0.0], [xd_, yd_, 0.0]])

                c_mirror, d_mirror = mirror_points_line([[xc_, yc_, 0.0], [xd_, yd_, 0.0]], line_hor)
                lines.append([c_mirror, d_mirror])
                c_mirror, d_mirror = mirror_points_line([c_mirror, d_mirror], line_ver)
                lines.append([c_mirror, d_mirror])
                c_mirror, d_mirror = mirror_points_line([[xc_, yc_, 0.0], [xd_, yd_, 0.0]], line_ver)
                lines.append([c_mirror, d_mirror])

    pattern: Pattern = Pattern.from_lines(lines, delete_boundary_face=True)  # type: ignore
    return pattern


def create_ortho_pattern(
    xy_span=[[0.0, 10.0], [0.0, 10.0]],
    discretisation=[10, 10],
) -> Pattern:
    """Helper to construct a FormDiagram based on a simple orthogonal discretisation.

    Parameters
    ----------
    xy_span : list, optional
        List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
    discretisation : int, optional
        Set the density of the grid in x and y directions, by default 10
    fix : str, optional
        Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.
    """

    if isinstance(discretisation, int):
        discretisation = [discretisation, discretisation]
    # if discretisation[0] % 2 != 0 or discretisation[1] % 2 != 0:
    #     msg = "Warning!: discretisation of this form diagram has to be even."
    #     raise ValueError(msg)

    y1 = float(xy_span[1][1])
    y0 = float(xy_span[1][0])
    x1 = float(xy_span[0][1])
    x0 = float(xy_span[0][0])
    x_span = x1 - x0
    y_span = y1 - y0
    division_x = discretisation[0]
    division_y = discretisation[1]
    dx = float(x_span / division_x)
    dy = float(y_span / division_y)

    vertices = []
    faces = []

    for j in range(division_y + 1):
        for i in range(division_x + 1):
            xi = x0 + dx * i
            yi = y0 + dy * j
            vertices.append([xi, yi, 0.0])
            if i < division_x and j < division_y:
                p1 = j * (division_x + 1) + i
                p2 = j * (division_x + 1) + i + 1
                p3 = (j + 1) * (division_x + 1) + i + 1
                p4 = (j + 1) * (division_x + 1) + i
                face = [p1, p2, p3, p4, p1]
                faces.append(face)
                print(face)

    pattern: Pattern = Pattern.from_vertices_and_faces(vertices, faces)  # type: ignore
    return pattern


def create_parametric_pattern(
    xy_span=[[0.0, 10.0], [0.0, 10.0]],
    discretisation=10,
    lambd=0.5,
) -> Pattern:
    """Create a parametric form diagram based on the inclination lambda of the arches

    Parameters
    ----------
    xy_span : [[float, float], [float, float]], optional
        List with initial- and end-points of the vault, by default, by default [[0.0, 10.0], [0.0, 10.0]]
    discretisation : int, optional
        Set the density of the grid in x and y directions, by default 10
    lambd : float, optional
        Inclination of the arches in the diagram (0.0 will result in cross and 1.0 in fan diagrams), by default 0.5
    fix : str, optional
        Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.

    Notes
    ---------
        Diagram implemented after `N. A. Nodargi et al., 2022 <https://doi.org/10.1016/j.engstruct.2022.114878>`_.
    """
    if 0.0 > lambd or lambd > 1.0:
        raise ValueError("Lambda should be between 0.0 and 1.0")

    x_span = xy_span[0][1] - xy_span[0][0]
    y_span = xy_span[1][1] - xy_span[1][0]

    if abs(x_span - y_span) > 1e-6:
        y_span = x_span = 10.0
        x0, x1 = y0, y1 = 0.0, 10.0
    else:
        x0, x1 = xy_span[0][0], xy_span[0][1]
        y0, y1 = xy_span[1][0], xy_span[1][1]

    xc = (x1 + x0) / 2
    yc = (y1 + y0) / 2

    xc0 = x0 + x_span / 2
    yc0 = y0 + y_span / 2
    division_x = discretisation
    division_y = discretisation
    dx = float(x_span / division_x)
    dy = float(y_span / division_y)
    nx = int(division_x / 2)
    line_hor = [[x0, yc0, 0.0], [xc0, yc0, 0.0]]
    line_ver = [[xc0, y0, 0.0], [xc0, yc0, 0.0]]

    lines = []

    for i in range(nx + 1):
        j = i

        xa = xc
        ya = xc - dy * j
        xb = (ya - y0) * (1 - lambd)
        yb = (ya - y0) * (1 - lambd)

        if distance_point_point_xy([xa, ya], [xb, yb]):
            lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

            append_mirrored_lines([[xa, ya, 0.0], [xb, yb, 0.0]], lines, line_hor, line_ver)

        if i == 0:
            xa = x0
            ya = y0

            if distance_point_point_xy([xa, ya], [xb, yb]):
                lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

                append_mirrored_lines([[xa, ya, 0.0], [xb, yb, 0.0]], lines, line_hor, line_ver)

        xa = xc - dx * j
        xb = xc - dx * j
        ya = yc - dy * j
        yb = y0

        if distance_point_point_xy([xa, ya], [xb, yb]):
            lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

            append_mirrored_lines([[xa, ya, 0.0], [xb, yb, 0.0]], lines, line_hor, line_ver)

        xa = yc - dy * j
        ya = yc
        xb = (xa - x0) * (1 - lambd)
        yb = xb * (y1 - y0) / (x1 - x0) + y0

        if distance_point_point_xy([xa, ya], [xb, yb]):
            lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

            append_mirrored_lines([[xa, ya, 0.0], [xb, yb, 0.0]], lines, line_hor, line_ver)

        xa = xc - dx * j
        xb = x0
        ya = yc - dy * j
        yb = yc - dy * j

        if distance_point_point_xy([xa, ya], [xb, yb]):
            lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

            append_mirrored_lines([[xa, ya, 0.0], [xb, yb, 0.0]], lines, line_hor, line_ver)

    # clean_lines = split_intersection_lines(lines)



import math

from compas.geometry import intersection_line_line_xy
from compas_rv.datastructures import Pattern


def create_circular_radial_pattern(
    center=[5.0, 5.0],
    radius=5.0,
    discretisation=[8, 20],
    r_oculus=0.0,
    diagonal=False,
    partial_diagonal=False,
) -> Pattern:
    """Construct a circular radial FormDiagram with hoops equally spaced in plan.

    Parameters
    ----------
    center : list, optional
        Planar coordinates of the form-diagram [xc, yc], by default [5.0, 5.0]
    radius : float, optional
        Radius of the form diagram, by default 5.0
    discretisation : list, optional
        Number of hoops, and of parallels of the dome form diagram], by default [8, 20]
    r_oculus : float, optional
        Value of the radius of the oculus, if no oculus is present should be set to zero, by default 0.0
    diagonal : bool, optional
        Activate diagonal in the quads, by default False
    partial_diagonal : bool, optional
        Activate partial diagonal in the quads, by default False

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.

    """

    xc = center[0]
    yc = center[1]
    n_radial = discretisation[0]
    n_spikes = discretisation[1]
    theta = 2 * math.pi / n_spikes
    r_div = (radius - r_oculus) / n_radial
    lines = []

    for nr in range(n_radial + 1):
        for nc in range(n_spikes):
            if (r_oculus + nr * r_div) > 0.0:
                # Meridian Elements
                xa = xc + (r_oculus + nr * r_div) * math.cos(theta * nc)
                xb = xc + (r_oculus + nr * r_div) * math.cos(theta * (nc + 1))
                ya = yc + (r_oculus + nr * r_div) * math.sin(theta * nc)
                yb = yc + (r_oculus + nr * r_div) * math.sin(theta * (nc + 1))
                lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

            if nr <= n_radial - 1:
                # Radial Elements
                xa = xc + (r_oculus + nr * r_div) * math.cos(theta * nc)
                xb = xc + (r_oculus + (nr + 1) * r_div) * math.cos(theta * nc)
                ya = yc + (r_oculus + nr * r_div) * math.sin(theta * nc)
                yb = yc + (r_oculus + (nr + 1) * r_div) * math.sin(theta * nc)
                lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

    if diagonal:
        for nr in range(n_radial):
            for nc in range(n_spikes):
                if (r_oculus + nr * r_div) > 0.0:
                    # Meridian Element i
                    xa = xc + (r_oculus + nr * r_div) * math.cos(theta * nc)
                    xb = xc + (r_oculus + nr * r_div) * math.cos(theta * (nc + 1))
                    ya = yc + (r_oculus + nr * r_div) * math.sin(theta * nc)
                    yb = yc + (r_oculus + nr * r_div) * math.sin(theta * (nc + 1))

                    # Meridian Element i + 1
                    xa_ = xc + (r_oculus + (nr + 1) * r_div) * math.cos(theta * nc)
                    xb_ = xc + (r_oculus + (nr + 1) * r_div) * math.cos(theta * (nc + 1))
                    ya_ = yc + (r_oculus + (nr + 1) * r_div) * math.sin(theta * nc)
                    yb_ = yc + (r_oculus + (nr + 1) * r_div) * math.sin(theta * (nc + 1))

                    if partial_diagonal == "right":
                        if nc + 1 > n_spikes / 2:
                            lines.append([[xa, ya, 0.0], [xb_, yb_, 0.0]])
                        else:
                            lines.append([[xa_, ya_, 0.0], [xb, yb, 0.0]])
                    elif partial_diagonal == "left":
                        if nc + 1 > n_spikes / 2:
                            lines.append([[xa_, ya_, 0.0], [xb, yb, 0.0]])
                        else:
                            lines.append([[xa, ya, 0.0], [xb_, yb_, 0.0]])
                    elif partial_diagonal == "rotation":
                        lines.append([[xa, ya, 0.0], [xb_, yb_, 0.0]])
                    elif partial_diagonal == "straight":
                        midx, midy, _ = intersection_line_line_xy([[xa, ya], [xb_, yb_]], [[xa_, ya_], [xb, yb]])  # type: ignore
                        lines.append([[xa, ya, 0.0], [midx, midy, 0.0]])
                        lines.append([[midx, midy, 0.0], [xb_, yb_, 0.0]])
                        lines.append([[xa_, ya_, 0.0], [midx, midy, 0.0]])
                        lines.append([[midx, midy, 0.0], [xb, yb, 0.0]])
                    else:
                        midx = (xa + xa_ + xb + xb_) / 4
                        midy = (ya + ya_ + yb + yb_) / 4
                        lines.append([[xa, ya, 0.0], [midx, midy, 0.0]])
                        lines.append([[midx, midy, 0.0], [xb_, yb_, 0.0]])
                        lines.append([[xa_, ya_, 0.0], [midx, midy, 0.0]])
                        lines.append([[midx, midy, 0.0], [xb, yb, 0.0]])

    pattern: Pattern = Pattern.from_lines(lines, delete_boundary_face=True)  # type: ignore
    return pattern


def create_circular_radial_spaced_pattern(
    center=[5.0, 5.0],
    radius=5.0,
    discretisation=[8, 20],
    r_oculus=0.0,
    diagonal=False,
    partial_diagonal=False,
) -> Pattern:
    """Construct a circular radial FormDiagram with hoops not equally spaced in plan, but equally spaced with regards to the projection on a hemisphere.

    Parameters
    ----------
    center : list, optional
        Planar coordinates of the form-diagram [xc, yc], by default [5.0, 5.0]
    radius : float, optional
        Radius of the form diagram, by default 5.0
    discretisation : list, optional
        Number of hoops, and of parallels of the dome form diagram], by default [8, 20]
    r_oculus : float, optional
        Value of the radius of the oculus, if no oculus is present should be set to zero, by default 0.0
    diagonal : bool, optional
        Activate diagonal in the quads, by default False
    partial_diagonal : bool, optional
        Activate partial diagonal in the quads, by default False

    Returns
    -------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.

    """
    xc = center[0]
    yc = center[1]
    n_radial = discretisation[0]
    n_spikes = discretisation[1]
    theta = 2 * math.pi / n_spikes
    r_div = (radius - r_oculus) / n_radial
    lines = []

    for nr in range(n_radial + 1):
        for nc in range(n_spikes):
            if (r_oculus + nr) > 0:
                # Meridian Elements
                xa = xc + (r_oculus + radius * math.cos((n_radial - nr) / n_radial * math.pi / 2)) * math.cos(theta * nc)
                xb = xc + (r_oculus + radius * math.cos((n_radial - nr) / n_radial * math.pi / 2)) * math.cos(theta * (nc + 1))
                ya = yc + (r_oculus + radius * math.cos((n_radial - nr) / n_radial * math.pi / 2)) * math.sin(theta * nc)
                yb = yc + (r_oculus + radius * math.cos((n_radial - nr) / n_radial * math.pi / 2)) * math.sin(theta * (nc + 1))
                lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

            if nr <= n_radial - 1:
                # Radial Elements
                xa = xc + (r_oculus + radius * math.cos((n_radial - nr) / n_radial * math.pi / 2)) * math.cos(theta * nc)
                xb = xc + (r_oculus + radius * math.cos((n_radial - (nr + 1)) / n_radial * math.pi / 2)) * math.cos(theta * nc)
                ya = yc + (r_oculus + radius * math.cos((n_radial - nr) / n_radial * math.pi / 2)) * math.sin(theta * nc)
                yb = yc + (r_oculus + radius * math.cos((n_radial - (nr + 1)) / n_radial * math.pi / 2)) * math.sin(theta * nc)
                lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

    if diagonal:
        for nr in range(n_radial):
            for nc in range(n_spikes):
                if (r_oculus + nr * r_div) > 0.0:
                    # Meridian Element i
                    xa = xc + (r_oculus + radius * math.cos((n_radial - nr) / n_radial * math.pi / 2)) * math.cos(theta * nc)
                    xb = xc + (r_oculus + radius * math.cos((n_radial - nr) / n_radial * math.pi / 2)) * math.cos(theta * (nc + 1))
                    ya = yc + (r_oculus + radius * math.cos((n_radial - nr) / n_radial * math.pi / 2)) * math.sin(theta * nc)
                    yb = yc + (r_oculus + radius * math.cos((n_radial - nr) / n_radial * math.pi / 2)) * math.sin(theta * (nc + 1))

                    # radius * math.cos((n_radial - (nr + 1)) / n_radial * math.pi / 2)

                    # Meridian Element i + 1
                    xa_ = xc + (r_oculus + radius * math.cos((n_radial - (nr + 1)) / n_radial * math.pi / 2)) * math.cos(theta * nc)
                    xb_ = xc + (r_oculus + radius * math.cos((n_radial - (nr + 1)) / n_radial * math.pi / 2)) * math.cos(theta * (nc + 1))
                    ya_ = yc + (r_oculus + radius * math.cos((n_radial - (nr + 1)) / n_radial * math.pi / 2)) * math.sin(theta * nc)
                    yb_ = yc + (r_oculus + radius * math.cos((n_radial - (nr + 1)) / n_radial * math.pi / 2)) * math.sin(theta * (nc + 1))
                    if partial_diagonal == "right":
                        if nc + 1 > n_spikes / 2:
                            lines.append([[xa, ya, 0.0], [xb_, yb_, 0.0]])
                        else:
                            lines.append([[xa_, ya_, 0.0], [xb, yb, 0.0]])
                    elif partial_diagonal == "left":
                        if nc + 1 > n_spikes / 2:
                            lines.append([[xa_, ya_, 0.0], [xb, yb, 0.0]])
                        else:
                            lines.append([[xa, ya, 0.0], [xb_, yb_, 0.0]])
                    elif partial_diagonal == "straight":
                        midx, midy, _ = intersection_line_line_xy([[xa, ya], [xb_, yb_]], [[xa_, ya_], [xb, yb]])  # type: ignore
                        lines.append([[xa, ya, 0.0], [midx, midy, 0.0]])
                        lines.append([[midx, midy, 0.0], [xb_, yb_, 0.0]])
                        lines.append([[xa_, ya_, 0.0], [midx, midy, 0.0]])
                        lines.append([[midx, midy, 0.0], [xb, yb, 0.0]])
                    else:
                        midx = (xa + xa_ + xb + xb_) / 4
                        midy = (ya + ya_ + yb + yb_) / 4
                        lines.append([[xa, ya, 0.0], [midx, midy, 0.0]])
                        lines.append([[midx, midy, 0.0], [xb_, yb_, 0.0]])
                        lines.append([[xa_, ya_, 0.0], [midx, midy, 0.0]])
                        lines.append([[midx, midy, 0.0], [xb, yb, 0.0]])

    pattern: Pattern = Pattern.from_lines(lines, delete_boundary_face=True)  # type: ignore
    return pattern


def create_circular_spiral_pattern(
    center=[5.0, 5.0],
    radius=5.0,
    discretisation=[8, 20],
    r_oculus=0.0,
) -> Pattern:
    """Construct a circular radial FormDiagram with hoops not equally spaced in plan, but equally spaced with regards to the projection on a hemisphere.

    Parameters
    ----------
    center : list, optional
        Planar coordinates of the form-diagram [xc, yc], by default [5.0, 5.0]
    radius : float, optional
        Radius of the form diagram, by default 5.0
    discretisation : list, optional
        Number of hoops, and of parallels of the dome form diagram], by default [8, 20]
    r_oculus : float, optional
        Value of the radius of the oculus, if no oculus is present should be set to zero, by default 0.0

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.

    """
    xc = center[0]
    yc = center[1]
    n_radial = discretisation[0]
    n_spikes = discretisation[1]
    theta = 2 * math.pi / n_spikes
    r_div = (radius - r_oculus) / n_radial
    lines = []

    for nr in range(n_radial + 1):
        for nc in range(n_spikes):
            if nr > 0.0:  # This avoid the center...
                if nr % 2 == 0:
                    # Diagonal to Up
                    xa = xc + (r_oculus + nr * r_div) * math.cos(theta * nc)
                    xb = xc + (r_oculus + (nr - 1) * r_div) * math.cos(theta * (nc + 1 / 2))
                    ya = yc + (r_oculus + nr * r_div) * math.sin(theta * nc)
                    yb = yc + (r_oculus + (nr - 1) * r_div) * math.sin(theta * (nc + 1 / 2))

                    # Diagonal to Down
                    xa_ = xc + (r_oculus + nr * r_div) * math.cos(theta * nc)
                    xb_ = xc + (r_oculus + (nr - 1) * r_div) * math.cos(theta * (nc - 1 / 2))
                    ya_ = yc + (r_oculus + nr * r_div) * math.sin(theta * nc)
                    yb_ = yc + (r_oculus + (nr - 1) * r_div) * math.sin(theta * (nc - 1 / 2))

                    lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])
                    lines.append([[xa_, ya_, 0.0], [xb_, yb_, 0.0]])
                else:
                    # Diagonal to Up
                    xa = xc + (r_oculus + nr * r_div) * math.cos(theta * (nc + 1 / 2))
                    xb = xc + (r_oculus + (nr - 1) * r_div) * math.cos(theta * (nc + 1))
                    ya = yc + (r_oculus + nr * r_div) * math.sin(theta * (nc + 1 / 2))
                    yb = yc + (r_oculus + (nr - 1) * r_div) * math.sin(theta * (nc + 1))

                    # Diagonal to Down
                    xa_ = xc + (r_oculus + nr * r_div) * math.cos(theta * (nc + 1 / 2))
                    xb_ = xc + (r_oculus + (nr - 1) * r_div) * math.cos(theta * (nc))
                    ya_ = yc + (r_oculus + nr * r_div) * math.sin(theta * (nc + 1 / 2))
                    yb_ = yc + (r_oculus + (nr - 1) * r_div) * math.sin(theta * (nc))

                    lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])
                    lines.append([[xa_, ya_, 0.0], [xb_, yb_, 0.0]])
                if nr == n_radial:
                    xa = xc + (r_oculus + nr * r_div) * math.cos(theta * nc)
                    xb = xc + (r_oculus + nr * r_div) * math.cos(theta * (nc + 1))
                    ya = yc + (r_oculus + nr * r_div) * math.sin(theta * nc)
                    yb = yc + (r_oculus + nr * r_div) * math.sin(theta * (nc + 1))
                    lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])
            if nr == 0 and r_oculus > 0.0:
                # If oculus, this will be the compression ring
                xa = xc + (r_oculus) * math.cos(theta * (nc))
                xb = xc + (r_oculus) * math.cos(theta * (nc + 1))
                ya = yc + (r_oculus) * math.sin(theta * (nc))
                yb = yc + (r_oculus) * math.sin(theta * (nc + 1))
                lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

    pattern: Pattern = Pattern.from_lines(lines, delete_boundary_face=True)  # type: ignore
    return pattern
